from direct.showbase.DirectObject import DirectObject
from pandac.PandaModules import NodePath,LineSegs
from direct.task.Task import Task
from direct.actor.Actor import Actor

import math

import Meshes

##############################################################################

class Avatars(DirectObject):
    
    #=========================================================================
    """initialize"""
    def __init__(self,ancestor=None):
        print "____________________________________________________"
        print "Class Avatars"
        self.ancestor=ancestor
        
        #----------------------
        #initialize basic props
        self.avList=[]
        self.avNum=0
        self.av=None
        #----------------------
        #add 4 avatars
        for n in range(0,4):
            self.addAvatar(path="../_models/avatars/ralph/",model="ralph",skin="black.jpg",name="avatar"+str(n),pos=None,dz=.03)
        #----------------------    
        #select initial avatar
        self.select(self.avNum)
        #set camera on initial avatar
        if self.ancestor.camera!=None:
            self.ancestor.camera.target.setPos(self.av.actor.getX(),self.av.actor.getY(),
                                               self.av.actor.getZ()+self.ancestor.camera.dz)
            base.camera.lookAt(self.ancestor.camera.target)
        
        #----------------------
        #enable key controls
        self.accept('space',self.selectNext)
        self.accept('c', self.center)
        self.accept('r', self.toggleRun)
        
    #=========================================================================
    def addAvatar(self,path,model,skin,name,pos,dz):
        num=len(self.avList)
        self.avList.append(Avatar(ancestor=self.ancestor,
                                      path=path,model=model,skin=skin,
                                      name=name,num=num,
                                      size=0.5*1.5,dz=dz,pos=pos,rot=None))
    def select(self,num):
        self.avNum=num
        self.av=self.avList[self.avNum]
        self.ancestor.txt.setText(self.av.name)
        self.center()
    def selectNext(self):
        self.avNum+=1
        if self.avNum>len(self.avList)-1:
            self.avNum=0
        self.select(self.avNum)
    def center(self):
        if self.ancestor.camera!=None:
            self.ancestor.camera.isCentering=True
    def toggleRun(self):
        self.av.isRunning=not self.av.isRunning
        if self.av.isMoving==True: self.av.startAnim() 
        
        
##############################################################################

class Avatar():
    
    #=========================================================================
    """initialize"""
    def __init__(self,ancestor=None,path=None,model=None,skin=None,name="avatar",num=0,size=.75,dz=0.25,pos=None,rot=None):
        #print "_________________________"
        print "    Class Avatar",name,num
        self.ancestor=ancestor
        
        #----------------------------------------------------
        #init props
        #----------------------------------------------------    
        self.name=name
        self.num=num
        self.dz=dz
        #path props
        self.path=[]
        self.startCell=None
        self.endCell=None
        self.gotoCell=None
        self.pathCubes=[]
        self.line=None
        self.lineNode=None
        #movement props
        self.isMoving=False
        self.isRunning=False
        self.willMove=None
        
        #----------------------------------------------------                    
        #init actor
        #----------------------------------------------------    
        self.actor=Meshes.loadActor(parent=self.ancestor.root,
                         path=path+model,anim={"walk":path+model+"-walk","run":path+model+"-run"},
                         size=size,pos=pos,color=None,tex=path+skin) 
        
        #set actor props
        self.actor.setPlayRate(1.95, 'walk')
        self.actor.setPlayRate(1.25, 'run')
        self.actor.pose('walk', 5)  
        #locate actor in random cell
        Meshes.locateRandom(self.ancestor,self.actor) 
        z=self.ancestor.terrain.getElevation(self.actor.getX(),self.actor.getY())
        self.actor.setZ(z+self.dz) 
        #register the actor class in a python tag
        self.actor.setPythonTag('class',self)
        #make instance pickable for collisions as 'avatar' type
        self.ancestor.picker.makePickable(self.actor,'avatar')

        #----------------------------------------------------
        #initialize move task
        taskMgr.add(self.move,"moveAvatarTask")
    
    #=========================================================================
    def initMove(self):
        if self.path==None: return
        #init vars
        self.isMoving=True
        self.moves=0
        #set initial pos
        self.lastx=self.actor.getX()
        self.lasty=self.actor.getY()
        #set initial cell to walkable 
        self.ancestor.grid.setCellWalkable(self.startCell[0],self.startCell[1],1)
        #set end cell unwalkable
        self.ancestor.grid.setCellWalkable(self.endCell[0],self.endCell[1],0)   
        #start animation
        self.startAnim()
    
    #=========================================================================
    def startAnim(self):
        #start animation
        if self.isRunning==True:
            self.actor.loop('run')
        else:
            self.actor.loop('walk')
        
    def stopAnim(self):
        self.actor.pose('walk', 5)
    
    #=========================================================================
    def drawPathLine(self,locs):
        #currently disabled: still have to find a way to stick the line 
        #perfectly to the terrain
        a=0
        ## #remove last line
        ## if self.lineNode!=None: self.lineNode.removeNode()
        ## #create lineSegs
        ## self.line = LineSegs()
        ## self.line.setThickness(100.0)
        ## self.line.moveTo(locs[len(locs)-1][0],locs[len(locs)-1][1],locs[len(locs)-1][2])
        ## for i in range(len(locs)):  
            ## n= len(locs)-i-1
            ## self.line.drawTo(locs[n][0],locs[n][1],locs[n][2]+self.dz)
        ## self.line.drawTo(self.actor.getX(),self.actor.getY(),self.actor.getZ())
        ## #create line node
        ## self.line.setThickness(2)
        ## node = self.line.create()
        ## self.lineNode = NodePath(node)
        ## self.lineNode.reparentTo(self.ancestor.root)
        ## self.lineNode.setColor(0,0,0,0)
        
    #=========================================================================
    def move(self, task):
        
        if self.isMoving==True:
            if len(self.path)>0:
                #-------------------------
                #move towards next cell
                #-------------------------
                #set location of next cell
                x=self.path[0][0]#+.5
                y=self.path[0][1]#+.5
                #set increments from last cell to next cell
                cdx=(x-self.lastx)
                cdy=(y-self.lasty)
                d=math.sqrt(cdx**2+cdy**2)
                #set vel
                vel=.1
                arrive=.05
                if self.isRunning==True:
                    vel=vel*2.0
                    arrive=.1
                #set increments from actor to next cell
                dx=(cdx/d)*vel
                dy=(cdy/d)*vel
                #orient actor towards next cell
                self.actor.lookAt(x,y,0+self.dz)
                self.actor.setP(0)
                self.actor.setH(self.actor.getH()+180)
                #move actor towards next cell
                self.actor.setX(self.actor.getX()+dx)
                self.actor.setY(self.actor.getY()+dy)
                #get terrain elevation at current xy
                z=self.ancestor.terrain.getElevation(self.actor.getX(),self.actor.getY())
                self.actor.setZ(z+self.dz) 
                
                self.drawPathLine(self.path)
                #-------------------------
                #arrive to next cell
                #-------------------------
                if(math.fabs(self.actor.getX()-x)<arrive and math.fabs(self.actor.getY()-y)<arrive):
                    #snap actor to cell
                    #cx,cy,cz=self.ancestor.terrain.getCellPos(self.actor.getX(),self.actor.getY())
                    self.actor.setPos(x,y,z+self.dz) 
                    #set avatar cell
                    self.cell=self.path[0]
                    #delete old path segment and old pathCube
                    del self.path[0]
                    self.pathCubes[self.moves].removeNode()
                    #actualize props
                    self.lastx=x
                    self.lasty=y
                    self.moves+=1
                    #if willMove enabled: abort movement and search new path
                    if self.willMove!=None:
                        self.stopAnim()
                        self.isMoving=False
                        self.path=[]
                        self.choosePath(self.willMove)  
            else:
                #-------------------------
                #end of movement
                #-------------------------
                self.stopAnim()
                self.isMoving=False
                self.endCell=None
                  
        # continue task
        return Task.cont
    
    #=========================================================================
    def choosePath(self,pickedPoint):
        #get end cell
        end=(int(math.floor(pickedPoint[0])),int(math.floor(pickedPoint[1])))
        #get start cell
        start=(int(math.floor(self.actor.getX())),int(math.floor(self.actor.getY())))
        self.startCell=start
        #set last end cell walkable
        if self.endCell!=None: self.ancestor.grid.setCellWalkable(self.endCell[0],self.endCell[1],1) 
        #set current cell unwalkable
        self.ancestor.grid.setCellWalkable(self.startCell[0],self.startCell[1],0) 
        #choose pathfind or walk
        if self.gotoCell!=end:
            #find path with pathfinder
            if self.ancestor.grid.data[end[0]][end[1]]['walkable']>0: self.path=self.ancestor.grid.pathFinder(self,start,end)
            self.endCell=end
            self.gotoCell=end
            self.willMove=None
            if self.path!=[]: self.ancestor.txt.setText(self.name+" path to "+str(self.endCell))
        else:
            #move avatar through path
            self.initMove()
            self.gotoCell=None
            #trace goto message
            if self.path!=[]: self.ancestor.txt.setText(self.name+" going to "+str(self.endCell))
            
################################################################################ 
        