
from direct.showbase.DirectObject import DirectObject
from pandac.PandaModules import NodePath,Vec3,Point3
from direct.task.Task import Task

import Meshes

##############################################################################

class Camera(DirectObject):
    
    #=========================================================================
    """initialize"""
    def __init__(self,ancestor=None,showTarget=False):
        print "____________________________________________________"
        print "Class Camera"
        self.ancestor=ancestor
        
        #------------------------------------------------------------
        #initialize basic props
        #---
        self.distMax=128
        self.distMin=4
        self.dist=self.distMin+4
        self.dz=.5
        center=(len(self.ancestor.terrain.data)/2,len(self.ancestor.terrain.data[0])/2,self.dz)
        #---
        self.iVel=.25
        self.isRotating=False
        self.rotVel=0.05
        self.isZooming=False
        self.zoomVel=2.0
        self.isCentering=False
        #------------------------------------------------------------
        #initialize camera target
        self.target=NodePath("cameraTarget")
        self.target.reparentTo(self.ancestor.root)
        if showTarget==True:
            self.target=Meshes.loadModel(parent=self.ancestor.root,
                                         path="../_models/shapes/cube",name="cameraTarget",size=0.1,pos=center,color=None)
        #------------------------------------------------------------
        #initialize camera
        base.disableMouse()
        base.camLens.setFov(50)
        base.camera.reparentTo(self.ancestor.root)
        d=8
        base.camera.setPos(self.target.getX()-d,self.target.getY()-d,self.target.getZ()+d)
        base.camera.lookAt(self.target)
        
        #------------------------------------------------------------
        #initialize mouse controls
        self.accept('mouse2', self.middleMouseDown, [])
        self.accept('mouse2-up', self.middleMouseUp, [])
        self.accept('wheel_down', self.wheelMouseDown, [])
        self.accept('wheel_up', self.wheelMouseUp, [])
        #------------------------------------------------------------
        #initialize camera task
        taskMgr.add(self.update,"updateCameraTask")
    
    #=========================================================================
    def middleMouseDown(self):
        #print "middleMouseDown"
        self.mPos=[base.win.getPointer(0).getX(),base.win.getPointer(0).getY()]
        self.isRotating=True
    def middleMouseUp(self):
        #print "middleMouseUp"
        self.isRotating=False
    def wheelMouseDown(self):
        #print "wheelMouseDown"
        self.isZooming=True
        self.dist+=self.zoomVel
        if self.dist>self.distMax: self.dist=self.distMax  
    def wheelMouseUp(self):
        #print "wheelMouseUp"
        self.isZooming=False
        self.dist-=self.zoomVel
        if self.dist<self.distMin: self.dist=self.distMin
        
    #=========================================================================
    def update(self, task):


        #point camera at target before any changes
        base.camera.lookAt(self.target)
        #---------------------------------------------------------------------
        
        #rotate camera
        #----------------------------
        if (self.isRotating==True):
            #get mouse increments
            newPos=[base.win.getPointer(0).getX(),base.win.getPointer(0).getY()]
            d=[newPos[0]-self.mPos[0],newPos[1]-self.mPos[1]]
            #interpolate mouse last pos to new pos
            self.mPos[0]+=d[0]*.5
            self.mPos[1]+=d[1]*.5
            #rotate camera using x vector (left/right)
            camright = base.camera.getNetTransform().getMat().getRow3(0)
            camright.normalize()
            base.camera.setPos(base.camera.getPos() - camright*d[0]*self.rotVel)
            #rotate camera using z vector (up/down)
            camup = base.camera.getNetTransform().getMat().getRow3(2)
            camup.normalize()
            base.camera.setPos(base.camera.getPos() + camup*d[1]*self.rotVel/2.5)
        
        #interpolate target to avatar
        #----------------------------
        if self.isCentering==True:
            if self.ancestor.avatars!=None:
                av=self.ancestor.avatars.av
                vec=(self.target.getPos()-av.actor.getPos()-Vec3(0,0,self.dz))*self.iVel
                self.target.setPos(self.target.getPos() - vec)
                base.camera.setPos(base.camera.getPos() - vec)
        #interpolate camera to target
        #----------------------------
        camvec = self.target.getPos() - base.camera.getPos()
        camdist = camvec.length()
        camvec.normalize()
        base.camera.setPos(base.camera.getPos() + camvec*(camdist-self.dist)*self.iVel)
        
        #camera collisions
        #----------------------------
        x=base.camera.getX()
        y=base.camera.getY()
        if x>=.5 and y>=.5 and x<=len(self.ancestor.terrain.data)-1.5 and y<=len(self.ancestor.terrain.data)-1.5:
            z=self.ancestor.terrain.getCellPos(base.camera.getX(),base.camera.getY())[2]+self.dz+1
        else:
            z=self.dz+1
        if base.camera.getZ()-z<0:
            base.camera.setZ(z)
        
        #---------------------------------------------------------------------   
        #point camera at target after all changes
        base.camera.lookAt(self.target)

        return Task.cont
        
##############################################################################