title="Retro World v.1.0"

"""
Retro World is a tiled based game prototype by Snaptothegrid.
At the moment creates a random tiny environment and some avatars to move around

Game Controls:
- Mouse click on terrain:       find path to cell, goto cell if clicked again
- Mouse right-click on terrain: cell info
- Mouse click on avatar:        select avatar
- Mouse Middle button:          rotate camera
- Mouse wheel:                  zoom camera
- Space:                        select next avatar
- r:                            toggle avatar walk/run state

Debug Controls:
- w:                            toggle wireframe
- p:                            take screenshot
- t:                            toggles texture enabled/disabled

Modules:
- Main: Initializes world and setup
- Meshes: functions for model management (load, resize, locate...)
- Map: generates random heightmap and colormap images
- Terrain: generates a random multitextured terrain mesh. can be set to retro or normal mode
- Grid: generates an abstract grid of tiles and manages pathfinding
- Camera: manages camera movement (Middle mouse: rotate camera, MouseWheel: zoom)
- Picker: Implements mouse collision detection on terrain and avatars
- Lights: adds some basic lights and fog to the environment
- Obstacles: add trees, plants, crates and barrels randomly
- Avatars: 
    -Avatars: adds some avatars to the world 
    -Avatar:  holds each avatar data and methods -> creation, selection, movement

Next version:
- fix getElevation algorythm: is not perfect in retro mode
- some very basic shadow system
- implement other creatures
- implement initial game mechanics

Thanks pro-soft for helping with comments and panda comunity in general!
Enjoy!

"""

#setup config file
from pandac.PandaModules import loadPrcFileData

loadPrcFileData( '', 'fullscreen 0')
loadPrcFileData( '', 'win-size 800 600' )
loadPrcFileData( '', 'win-origin 450 90' )
loadPrcFileData( '', 'frame-rate-meter-scale 0.035' )
loadPrcFileData( '', 'frame-rate-meter-side-margin 0.1' )
loadPrcFileData( '', 'show-frame-rate-meter 1' )

loadPrcFileData( '', 'load-display pandagl' )
###loadPrcFileData( '', 'load-display pandadx9' )
loadPrcFileData( '', 'notify-level-util error' )
loadPrcFileData( '', 'window-title '+title )

#loadPrcFileData( '', 'auto-generate-mipmap 1' )
#loadPrcFileData( '', 'egg-flat-shading 1' )
#loadPrcFileData("", "sync-video 0")
#loadPrcFileData("", "even-animation 1")

loadPrcFileData("", "interpolate-frames 1")
loadPrcFileData("", "clock-mode limited")
loadPrcFileData("", "clock-frame-rate 25")
 
#import modules
import direct.directbase.DirectStart
from direct.showbase.DirectObject import DirectObject
from direct.gui.OnscreenText import OnscreenText,TextNode
from pandac.PandaModules import NodePath
import sys
import Picker,Terrain,Grid,Camera,Lights,Obstacles,Avatars

##############################################################################

class Main(DirectObject):     
    #=========================================================================
    """initialize"""
    def __init__(self):
        print "____________________________________________________"
        print "Class Main"
        #-----------------------
        #Setup application
        #-----------------------
        self.keys()
        self.txt=self.info((-1.32,0.96),title)
        #-----------------------
        #Initialize classes
        #-----------------------
        self.grid=None
        self.camera=None
        self.avatars=None
        #Initialize World
        self.root=NodePath("rootMain")
        self.root.reparentTo(render)
        #Initialize Picker
        self.picker=Picker.Picker(self)
        #Initialize Terrain
        self.terrain=Terrain.Terrain(self)
        #Initialize Grid
        self.grid=Grid.Grid(self,self.terrain.data)
        #Initialize Camera
        self.camera=Camera.Camera(self,False) 
        #Initialize Lights
        self.lights=Lights.Lights(self,lightsOn=True,showLights=False)
        #Initialize Obstacles 
        self.obstacles=Obstacles.Obstacles(self)        
        #Initialize Avatars
        self.avatars=Avatars.Avatars(self)      
        #-----------------------
        #analyze scene
        #render.analyze()

    #=========================================================================
    """info"""
    def info(self,pos, msg):
        self.font = loader.loadFont('../_media/fonts/OCR.otf')
        return OnscreenText(font=self.font,text=msg, style=1, fg=(1,1,1,1),
                            pos=pos, align=TextNode.ALeft, scale = .035, mayChange=True)
    #=========================================================================
    """keys"""
    def keys(self):
        self.accept('w',self.toggleWireFrame)
        self.accept('t',self.toggleTexture)
        self.accept('s',self.snapShot)
        self.accept('escape',sys.exit)
    def snapShot(self):
        base.screenshot("Snapshot")
    def toggleWireFrame(self):
        base.toggleWireframe()
    def toggleTexture(self):
        base.toggleTexture()
    
##############################################################################

main=Main()
run()