from pandac.PandaModules import NodePath
import os,random
import Meshes

##############################################################################

class Obstacles:
    #=========================================================================
    """initialize"""
    def __init__(self,ancestor):
        print "____________________________________________________"
        print "Class Obstacles"   
        self.ancestor=ancestor
        
        #--------------------------------------
        #create obstacles nodepath
        self.root=NodePath('obstaclesRoot')
        self.root.reparentTo(self.ancestor.root)
        
        #set vars
        self.path='../_models/obstacles/'
        
        #--------------------------------------
        #create obstacles
        
        #trees
        dir='trees/'
        data=[{'name':'palm-v1',   'size':2.0, 'dz':-.5,  'tex': None},
              {'name':'muddytree', 'size':1.5, 'dz':-.1,    'tex': None},
              {'name':'tree',     'size':1.5, 'dz':-.1,    'tex': 4}]
        self.createObstacles(dir,data,64/len(data))
        
        #plants 
        dir='plants/'   
        data=[{'name':'farn_1',      'size':1.0, 'dz':0, 'tex': None},
              {'name':'pflanze',     'size':1.0, 'dz':0, 'tex': None},
              {'name':'plants3',     'size':1.0, 'dz':0, 'tex': None},
              {'name':'grass',       'size':1.0, 'dz':0, 'tex': None},
              {'name':'leafy_plant', 'size':1.0, 'dz':0, 'tex': None},
              {'name':'shrubbery',   'size':1.0, 'dz':0, 'tex': None},
              {'name':'shrubbery2',  'size':.5,  'dz':0, 'tex': None},
              {'name':'plants1',     'size':1.0, 'dz':0, 'tex': None}]
        self.createObstacles(dir,data,128/len(data))
               
        #containers
        dir='containers/'
        data=[{'name':'crate', 'size':.75, 'dz':0, 'tex': 4},
              {'name':'barrel','size':.6, 'dz':0, 'tex': 7},
              {'name':'woodenbarrel',  'size':.6, 'dz':0, 'tex': None}]
        self.createObstacles(dir,data,32/len(data))             
        
        #--------------------------------------
        #flatten all obstacles
        self.root.flattenLight()
    
    #=========================================================================
    def createObstacles(self,dir,data,max):
        for n in range(len(data)):
            path=self.path+dir+data[n]['name']+"/"
            model=Meshes.loadModel(parent=self.root,path=path+data[n]['name'],
                        name=data[n]['name'],size=data[n]['size']*1.25,pos=None,color=None,tex=None)
            self.createInstances(model,max,path,data[n])
            model.removeNode()      
    
    #=========================================================================    
    def createInstances(self,model,max,path,data):
        #set possible textures
        if data['tex']!=None:
            textures = [ loader.loadTexture(path+data['name']+"%d.jpg" % i) for i in range(data['tex']) ]
        #copy model source
        for i in range(max):
            instance=model.copyTo(model.getParent()) 
            Meshes.locateRandom(self.ancestor,instance,data['dz'])
            if data['tex']!=None:
                tx=textures[random.randint(0, data['tex']-1)]#print "***",tx
                instance.setTexture(tx, 1)
                

##############################################################################
    