from direct.showbase import DirectObject
from pandac.PandaModules import CollisionTraverser,CollisionHandlerQueue,CollisionNode,CollisionRay,GeomNode
import math
import Meshes

class Picker(DirectObject.DirectObject):
    def __init__(self,ancestor=None):
        print "____________________________________________________"
        print "Class Picker"
        self.ancestor=ancestor
        #create traverser
        base.cTrav = CollisionTraverser()
        #create collision ray
        self.createRay(self,base.camera,name="mouseRay",show=True)
        #initialize mousePick
        self.accept('mouse1-up', self.mousePick, [1,self.queue])
        #initialize mouseRightPick
        self.accept('mouse3', self.mousePick, [3,self.queue])
    
    #===========================================================================
    def mouseRight(self,pickedObj,pickedPoint):
        if pickedObj==None: return
        #get cell from pickedpoint
        cell=(int(math.floor(pickedPoint[0])),int(math.floor(pickedPoint[1])))
        #trace updated info about the cell
        self.ancestor.txt.setText(str(self.ancestor.grid.data[cell[0]][cell[1]]))  
            
    #===========================================================================
    def mouseLeft(self,pickedObj,pickedPoint):
        if pickedObj==None: return
        #actions for AVATAR: select
        #------------------------------
        if pickedObj.getTag('pickable')=='avatar':
            av=pickedObj.getPythonTag('class')
            self.ancestor.avatars.select(av.num)
        #actions for TERRAIN: find path 
        #------------------------------
        if pickedObj.getTag('pickable')=='terrain': 
            av=self.ancestor.avatars.av
            if av.isMoving==False:
                av.choosePath(pickedPoint) 
            else:
                av.willMove=pickedPoint

    #=========================================================================== 
    """mouse pick"""    
    def mousePick(self, but, queue):
        #get mouse coords
        if base.mouseWatcherNode.hasMouse()==False: return
        mpos=base.mouseWatcherNode.getMouse()
        #locate ray from camera lens to mouse coords
        self.rayNP.reparentTo(base.camera)
        self.ray.setFromLens(base.camNode, mpos.getX(),mpos.getY())
        #get collision: picked obj and point
        pickedObj,pickedPoint=self.getCollision(queue)
        #call appropiate mouse function (left or right)
        if but==1:self.mouseLeft(pickedObj,pickedPoint)
        if but==3:self.mouseRight(pickedObj,pickedPoint)
        
    #===========================================================================
    """Returns the picked nodepath and the picked 3d point"""
    def getCollision(self, queue):
        #do the traverse
        base.cTrav.traverse(render)
        #process collision entries in queue
        if queue.getNumEntries() > 0:
            queue.sortEntries()
            for i in range(queue.getNumEntries()):
                collisionEntry=queue.getEntry(i)
                pickedObj=collisionEntry.getIntoNodePath()
                #iterate up in model hierarchy to found a pickable tag
                parent=pickedObj.getParent()
                for n in range(4):
                    if parent.getTag('pickable')!="" or parent==render: break
                    parent=parent.getParent()
                #return appropiate picked object
                if parent.getTag('pickable')!="":
                    pickedObj=parent
                    pickedPoint = collisionEntry.getSurfacePoint(pickedObj)
                    #pickedNormal = collisionEntry.getSurfaceNormal(self.ancestor.worldNode)
                    #pickedDistance=pickedPoint.lengthSquared()#distance between your object and the collision
                    return pickedObj,pickedPoint         
        return None,None
    
    #===========================================================================
    """sets nodepath pickable state"""
    def makePickable(self,newObj,tag='true'):
        newObj.setTag('pickable',tag)
        #print "pickable: ",newObj,"as",tag
    
    #===========================================================================
    """creates a ray for detecting collisions"""
    def createRay(self,obj,ent,name,show=False,x=0,y=0,z=0,dx=0,dy=0,dz=-1):
        #create queue
        obj.queue=CollisionHandlerQueue()
        #create ray  
        obj.rayNP=ent.attachNewNode(CollisionNode(name))
        obj.ray=CollisionRay(x,y,z,dx,dy,dz)
        obj.rayNP.node().addSolid(obj.ray)
        obj.rayNP.node().setFromCollideMask(GeomNode.getDefaultCollideMask())
        base.cTrav.addCollider(obj.rayNP, obj.queue) 
        if show: obj.rayNP.show()
        
     #===========================================================================

        